
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#ifndef _HCL_GPUGEMMOOC_H_
#define _HCL_GPUGEMMOOC_H_

/*--------------------------------------------------------*/

#include <cublas_v2.h>

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

/**
 * GEMM operation:
 * C = alpha * op(A) * B + beta * op(C)
 */

/**
 * An opaque handle for GEMM operations.
 */
template <typename T>
struct gemmHandle;

/*--------------------------------------------------------*/

/**
 * Create a handle for GEMM operations.
 *
 * @param M Number of rows of op(A) and C
 * @param N Number of columns of op(B) and C
 * @param K Number of columns of op(A) and number of rows of op(B)
 * @param start The start iteration number of outer algorithm
 *              For example: SUMMA
 *              Usually 0.
 * @param end The number of iterations of the outer algorithm
 *            For example: SUMMA
 * @param deviceId The GPU on which to compute the GEMM.
 *                 If parameter 'numDevices' is > 1, then
 *                 it represents the starting GPU of the 
 *                 list of GPUs to use for GEMM
 * @param numDevices The number of devices to use of GEMM
 * @param overlap Overlap computation and data transfer
 *                Uses CUDA streams
 *                Recommended to use 1
 * @param algorithm Different algorithms
 * @param handle Output GEMM handle
 * @return 0 on success; Otherwise cuda/cublas error is
 *           returned
 */
template <typename T>
int
create(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<T>** handle
);

/*--------------------------------------------------------*/

/**
 * Destroy the handle for GEMM operations.
 * @param handle GEMM handle to destroy
 * @return 0 on success
 */
template <typename T>
int
destroy(gemmHandle<T>* handle);

/*--------------------------------------------------------*/

/**
 * Execute GEMM.
 *
 * Five types are supported (T):
 * <type>:
 * -------
 * float
 * double
 * cuComplex
 * cuDoubleComplex
 * __half
 *
 * @param handle GEMM handle
 * @param transa operation op(A) that is non- or (conj.) transpose.
 * @param transb operation op(B) that is non- or (conj.) transpose.
 * @param M Number of rows of op(A) and C
 * @param N Number of columns of op(B) and C
 * @param K Number of columns of op(A) and number of rows of op(B)
 * @param alpha <type> scalar used for multiplication
 * @param A <type> array of dimensions lda x k with lda>=max(1,m) if transa == CUBLAS_OP_N and lda x m with lda>=max(1,k) otherwise.
 * @param lda leading dimension of two-dimensional array used to store the matrix A.
 * @param B <type> array of dimension ldb x n with ldb>=max(1,k) if transa == CUBLAS_OP_N and ldb x k with ldb>=max(1,n) otherwise.
 * @param ldb leading dimension of two-dimensional array used to store matrix B.
 * @param beta <type> scalar used for multiplication. If beta==0, C does not have to be a valid input.
 * @param C <type> array of dimensions ldc x n with ldc>=max(1,m).
 * @param ldc leading dimension of a two-dimensional array used to store the matrix C.
 * @return 0 on success; Otherwise cuda/cublas error is
 *           returned
 *
 * @note There are current restrictions since these were
 * written as sub-routines for parallel outer algorithms such as
 * SUMMA.
 * 1). M > K
 * 2). N > K
 * 3). M and N must be multiple of K
 */
template <typename T>
int gemm(gemmHandle<T>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const T alpha,
         const T *A, const int lda, const T *B, const int ldb,
         const T beta, T *C, const int ldc
);

/*--------------------------------------------------------*/

}

/*--------------------------------------------------------*/

#endif /* _HCL_GPUGEMMOOC_H_ */

/*--------------------------------------------------------*/
