
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cuda_runtime.h>
#include "gemminternal.hpp"
#include "debug.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int gemm(gemmHandle<T>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const T alpha,
         const T *A, const int lda, const T *B, const int ldb,
         const T beta, T *C, const int ldc)
{
    if (handle->num_devices == 1)
    {
       switch (handle->kernel_type)
       {
         case FIT:
              return singleFit<T>(
                     handle,
                     transa, transb,
                     M, N, K, 
                     alpha,
                     A, lda, B, ldb,
                     beta, C, ldc
              );
              break;
         case EXCEED:
              {
                 switch (handle->algorithm)
                 {
                    case 0:
                      return singleExceed0<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                      );
                      break;
                    case 1:
                      return singleExceed1<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                      );
                      break;
                    default:
                      Debug::Error(
                         "No kernel available.\n");
                 }
              }
         default:
              Debug::Error(
                 "No kernel available.\n");
       }
    }

    switch (handle->kernel_type)
    {
      case FIT:
           return multigpuFit<T>(
                  handle,
                  transa, transb,
                  M, N, K, 
                  alpha,
                  A, lda, B, ldb,
                  beta, C, ldc
           );
           break;
      case EXCEED:
           {
              switch (handle->algorithm)
              {
                case 0:
                   return multigpuExceed_nd2_1<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                      );
                    break;
                case 1:
                   return multigpuExceed_np2_1<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                   );
                   break;
                case 2:
                   return multigpuExceed_nd2_2<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                   );
                   break;
                case 3:
                   return multigpuExceed_np2_2<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                   );
                   break;
                case 4:
                   return multigpuExceed_nd2_3<T>(
                          handle,
                          transa, transb,
                          M, N, K, 
                          alpha,
                          A, lda, B, ldb,
                          beta, C, ldc
                   );
                   break;
                default:
                   Debug::Error(
                      "No kernel available.\n");
              }
           }
           break;
      default:
           Debug::Error(
             "No kernel available.\n");
    }

    return 0;
}

/*--------------------------------------------------------*/

template
int gemm<double>(gemmHandle<double>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const double alpha,
         const double *A, const int lda, const double *B, const int ldb,
         const double beta, double *C, const int ldc);

template
int gemm<float>(gemmHandle<float>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const float alpha,
         const float *A, const int lda, const float *B, const int ldb,
         const float beta, float *C, const int ldc);

template
int gemm<cuComplex>(gemmHandle<cuComplex>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const cuComplex alpha,
         const cuComplex *A, const int lda, const cuComplex *B, const int ldb,
         const cuComplex beta, cuComplex *C, const int ldc);

template
int gemm<cuDoubleComplex>(gemmHandle<cuDoubleComplex>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const cuDoubleComplex alpha,
         const cuDoubleComplex *A, const int lda, const cuDoubleComplex *B, const int ldb,
         const cuDoubleComplex beta, cuDoubleComplex *C, const int ldc);

template
int gemm<__half>(gemmHandle<__half>* handle,
         cublasOperation_t transa, cublasOperation_t transb,
         const int M, const int N, const int K, const __half alpha,
         const __half *A, const int lda, const __half *B, const int ldb,
         const __half beta, __half *C, const int ldc);

/*--------------------------------------------------------*/

}

/*--------------------------------------------------------*/
