
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cstring>
#include "gemminternal.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int multigpuFit(gemmHandle<T> *handle, cublasOperation_t transa,
                cublasOperation_t transb, const int M, const int N, const int K,
                const T alpha, const T *A, const int lda, const T *B,
                const int ldb, const T beta, T *C, const int ldc) {
  int nd = handle->num_device_use;
  int w = N / K;  // if w = 1?
  int width[nd];
  int j;
  for (j = 0; j < nd; j++) {
    if (w % nd == 0) {
      width[j] = w / nd;
    } else {
      width[j] = (j < w % nd) ? (w / nd + 1) : (w / nd);
    }
  }

  // initialization
  int device_id = handle->device_id;
  if (handle->idx_outer == 0) {
    handle->dA = (T **)malloc(sizeof(T *) * nd);
    handle->dB = (T **)malloc(sizeof(T *) * nd);
    handle->dC = (T **)malloc(sizeof(T *) * nd);

    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // alloc memory
      CUDA_SAFE_CALL(cudaMalloc((void **)(&handle->dA[j]), sizeof(T) * M * K));
      CUDA_SAFE_CALL(
          cudaMalloc((void **)(&handle->dB[j]), sizeof(T) * K * width[j] * K));
      CUDA_SAFE_CALL(
          cudaMalloc((void **)(&handle->dC[j]), sizeof(T) * M * width[j] * K));
    }
    // C = alpha*AxB + beta*C
    device_id = handle->device_id;
    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // synchronous call, control is not returned to host before the device has
      // completed the requested task
      CUDA_SAFE_CALL(cudaMemcpy2D(handle->dC[j], sizeof(T) * width[j] * K,
                                  getBaddrFromIndex<T>(C, 0, j, NULL, width, N, K, 3),
                                  sizeof(T) * ldc, sizeof(T) * width[j] * K, M,
                                  cudaMemcpyHostToDevice));
    }
  }

  // send A and B[j], and do gemm
  device_id = handle->device_id;
  for (j = 0; j < nd; j++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // data transfer
    CUDA_SAFE_CALL(cudaMemcpy2D(handle->dA[j], sizeof(T) * K, A,
                                sizeof(T) * lda, sizeof(T) * K, M,
                                cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy2D(handle->dB[j], sizeof(T) * width[j] * K,
                                getBaddrFromIndex<T>(B, 0, j, NULL, width, N, K, 2),
                                sizeof(T) * ldb, sizeof(T) * width[j] * K, K,
                                cudaMemcpyHostToDevice));

    // Note: cublas is column-major, so have to swap matrices A and B
    CUBLAS_SAFE_CALL(cublasGemm<T>(
        handle->handles[j], transb,
        transa, width[j] * K, M, K, &alpha, handle->dB[j],
        width[j] * K, handle->dA[j], K, &beta, handle->dC[j], width[j] * K));
  }

  // index increases by 1
  handle->idx_outer++;

  // at the last iteration, copy result back, and free device memory, and
  // destroy handles
  if (handle->idx_outer == handle->num_iteration) {
    device_id = handle->device_id;
    for (j = 0; j < nd; j++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // synchronous call, control is not returned to host before the device has
      // completed the requested task
      CUDA_SAFE_CALL(
          cudaMemcpy2D(getBaddrFromIndex<T>(C, 0, j, NULL, width, N, K, 3), sizeof(T) * ldc,
                       handle->dC[j], sizeof(T) * width[j] * K,
                       sizeof(T) * width[j] * K, M, cudaMemcpyDeviceToHost));
      // free memory
      CUDA_SAFE_CALL(cudaFree(handle->dA[j]));
      handle->dA[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(handle->dB[j]));
      handle->dB[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(handle->dC[j]));
      handle->dC[j] = NULL;
    }
    free(handle->dA);
    handle->dA = NULL;
    free(handle->dB);
    handle->dB = NULL;
    free(handle->dC);
    handle->dC = NULL;
  }

  return 0;
}

/*--------------------------------------------------------*/

template int multigpuFit<double>(gemmHandle<double> *handle,
                                 cublasOperation_t transa,
                                 cublasOperation_t transb, const int M,
                                 const int N, const int K, const double alpha,
                                 const double *A, const int lda,
                                 const double *B, const int ldb,
                                 const double beta, double *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuFit<float>(gemmHandle<float> *handle,
                                cublasOperation_t transa,
                                cublasOperation_t transb, const int M,
                                const int N, const int K, const float alpha,
                                const float *A, const int lda, const float *B,
                                const int ldb, const float beta, float *C,
                                const int ldc);

/*--------------------------------------------------------*/

template int multigpuFit<cuComplex>(gemmHandle<cuComplex> *handle,
                                    cublasOperation_t transa,
                                    cublasOperation_t transb, const int M,
                                    const int N, const int K,
                                    const cuComplex alpha, const cuComplex *A,
                                    const int lda, const cuComplex *B,
                                    const int ldb, const cuComplex beta,
                                    cuComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuFit<cuDoubleComplex>(
    gemmHandle<cuDoubleComplex> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const cuDoubleComplex alpha, const cuDoubleComplex *A, const int lda,
    const cuDoubleComplex *B, const int ldb, const cuDoubleComplex beta,
    cuDoubleComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuFit<__half>(gemmHandle<__half> *handle,
                                 cublasOperation_t transa,
                                 cublasOperation_t transb, const int M,
                                 const int N, const int K, const __half alpha,
                                 const __half *A, const int lda,
                                 const __half *B, const int ldb,
                                 const __half beta, __half *C, const int ldc);

/*--------------------------------------------------------*/
}

/*--------------------------------------------------------*/
