
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cstring>
#include "gemminternal.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int singleFit(gemmHandle<T> *handle, cublasOperation_t transa,
              cublasOperation_t transb, const int M, const int N, const int K,
              const T alpha, const T *A, const int lda, const T *B,
              const int ldb, const T beta, T *C, const int ldc) {
  if (handle->idx_outer == 0) {
    handle->dA = (T **)malloc(sizeof(T *));
    handle->dB = (T **)malloc(sizeof(T *));
    handle->dC = (T **)malloc(sizeof(T *));

    CUDA_SAFE_CALL(cudaMalloc((void **)&handle->dA[0], sizeof(T) * M * K));
    CUDA_SAFE_CALL(cudaMalloc((void **)&handle->dB[0], sizeof(T) * K * N));
    CUDA_SAFE_CALL(cudaMalloc((void **)&handle->dC[0], sizeof(T) * M * N));

    // C = alpha*AxB + beta*C
    CUDA_SAFE_CALL(cudaMemcpy(handle->dC[0], C, sizeof(T) * M * N,
                              cudaMemcpyHostToDevice));
  }

  // transfer matrices A and B to device memory, note lda may change
  CUDA_SAFE_CALL(cudaMemcpy2D(handle->dA[0], sizeof(T) * K, A, sizeof(T) * lda,
                              sizeof(T) * K, M, cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy2D(handle->dB[0], sizeof(T) * N, B, sizeof(T) * ldb,
                              sizeof(T) * N, K, cudaMemcpyHostToDevice));

  // Note: cublas is column-major, so have to swap matrices A and B
  CUBLAS_SAFE_CALL(cublasGemm<T>(handle->handles[0], transb, transa, N, M, K,
                                 &alpha, handle->dB[0], N, handle->dA[0], K,
                                 &beta, handle->dC[0], N));

  CUDA_SAFE_CALL(cudaDeviceSynchronize());

  handle->idx_outer++;

  if (handle->idx_outer == handle->num_iteration) {
    CUDA_SAFE_CALL(cudaMemcpy(C, handle->dC[0], sizeof(T) * M * N,
                              cudaMemcpyDeviceToHost));
    CUDA_SAFE_CALL(cudaFree(handle->dA[0]));
    free(handle->dA);
    handle->dA = NULL;
    CUDA_SAFE_CALL(cudaFree(handle->dB[0]));
    free(handle->dB);
    handle->dB = NULL;
    CUDA_SAFE_CALL(cudaFree(handle->dC[0]));
    free(handle->dC);
    handle->dC = NULL;
  }

  return 0;
}

/*--------------------------------------------------------*/

template int singleFit<double>(gemmHandle<double> *handle,
                               cublasOperation_t transa,
                               cublasOperation_t transb, const int M,
                               const int N, const int K, const double alpha,
                               const double *A, const int lda, const double *B,
                               const int ldb, const double beta, double *C,
                               const int ldc);

/*--------------------------------------------------------*/

template int singleFit<float>(gemmHandle<float> *handle,
                              cublasOperation_t transa,
                              cublasOperation_t transb, const int M,
                              const int N, const int K, const float alpha,
                              const float *A, const int lda, const float *B,
                              const int ldb, const float beta, float *C,
                              const int ldc);

/*--------------------------------------------------------*/

template int singleFit<cuComplex>(gemmHandle<cuComplex> *handle,
                                  cublasOperation_t transa,
                                  cublasOperation_t transb, const int M,
                                  const int N, const int K,
                                  const cuComplex alpha, const cuComplex *A,
                                  const int lda, const cuComplex *B,
                                  const int ldb, const cuComplex beta,
                                  cuComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int singleFit<cuDoubleComplex>(
    gemmHandle<cuDoubleComplex> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const cuDoubleComplex alpha, const cuDoubleComplex *A, const int lda,
    const cuDoubleComplex *B, const int ldb, const cuDoubleComplex beta,
    cuDoubleComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int singleFit<__half>(gemmHandle<__half> *handle,
                               cublasOperation_t transa,
                               cublasOperation_t transb, const int M,
                               const int N, const int K, const __half alpha,
                               const __half *A, const int lda, const __half *B,
                               const int ldb, const __half beta, __half *C,
                               const int ldc);

/*--------------------------------------------------------*/
}

/*--------------------------------------------------------*/
