
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#ifndef _HCL_GEMMINTERNAL_H_
#define _HCL_GEMMINTERNAL_H_

/*--------------------------------------------------------*/

#include "gemm.hpp"
#include <cuda_runtime.h>
#include <cstdio>
#include <cstdlib>

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

#define OOC_GAMMA 0.9

/*--------------------------------------------------------*/

#define CUDA_SAFE_CALL(call) {                                           \
    cudaError_t result = call;                                           \
    if (result != cudaSuccess) {                                         \
      fprintf(stderr,                                                    \
              "Cuda error %d "                                           \
              "in file '%s' in line %i.\n",                              \
              result, __FILE__, __LINE__ );                              \
      exit(EXIT_FAILURE);                                                \
    }                                                                    \
}

/*--------------------------------------------------------*/

#define CUBLAS_SAFE_CALL( call ) {                                       \
    cublasStatus_t err = call;                                           \
    if( CUBLAS_STATUS_SUCCESS != err) {                                  \
        fprintf(stderr, "Cublas error %d in file '%s' in line %i.\n",    \
                err, __FILE__, __LINE__ );                               \
        exit(EXIT_FAILURE);                                              \
    }                                                                    \
}

/*--------------------------------------------------------*/

/**
 * Types of kernel
 */
enum KernelType {
   FIT = 0,      /**< Fits memory(s) of GPU(s) */
   EXCEED,       /**< Exceeds memory(s) of GPU(s) */
};

/*--------------------------------------------------------*/

/*! Internal GEMM reuse structure */
/*! 
 * when idx_outer is equal to num_iteration, it's the last step of the outer algorithm,
 * the intermediate results in device should be copied back to host, 
 * and free device memory when necessary.
 */
template <typename T>
struct gemmHandle {
  /*! device memory pointer, size is num_device_use */
  T** dA, **dB, **dC;
  /*! pinned memory for data transfers of C between host and device */
  // try to avoid conflicts with IB when allocating huge pinned memory for C
  // it will only allocate pinned memory of the memory capacuties of devices
  // use stream callback function to schedule data transfers
  T** pC;
  /*! pointers to host workspace WA, WB, and DMA workspace of WB*/
  T *workspace_WA, *workspace_WB;
  // somehow MPI_Bcast does not work on pinned memory, so workspace_WB is 
  T* dma_WB;
  /*! pointer to CUBLAS handles */
  cublasHandle_t* handles;
  /*! the index of iteration of the outer algorithm */
  int idx_outer;
  /*! the number of iterations of the outer algorithm */
  int num_iteration;
  /*! the type of kernels */
  KernelType kernel_type;
  /*! the number of devices that are actually used for computing */
  int num_device_use;
  /*! the starting device id */
  int device_id;
  /*! number of devices */
  int num_devices;
  /*! memory allocation mode */
  int memory_mode;
  /*! cuda gemm algorithms */
  int algorithm;
  /*! whether use concurrent data transfer and computation or not */
  int is_overlap;
  /*! Core memory size */
  size_t core_device_mem_size;
};

/*--------------------------------------------------------*/

template <typename T>
int singleFit(gemmHandle<T>* handle,
              cublasOperation_t transa, cublasOperation_t transb,
              const int M, const int N, const int K, const T alpha,
              const T *A, const int lda, const T *B, const int ldb,
              const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int singleExceed0(gemmHandle<T>* handle,
                  cublasOperation_t transa, cublasOperation_t transb,
                  const int M, const int N, const int K, const T alpha,
                  const T *A, const int lda, const T *B, const int ldb,
                  const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int singleExceed1(gemmHandle<T>* handle,
                  cublasOperation_t transa, cublasOperation_t transb,
                  const int M, const int N, const int K, const T alpha,
                  const T *A, const int lda, const T *B, const int ldb,
                  const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int singleExceed2(gemmHandle<T>* handle,
                  cublasOperation_t transa, cublasOperation_t transb,
                  const int M, const int N, const int K, const T alpha,
                  const T *A, const int lda, const T *B, const int ldb,
                  const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int singleExceed3(gemmHandle<T>* handle,
                  cublasOperation_t transa, cublasOperation_t transb,
                  const int M, const int N, const int K, const T alpha,
                  const T *A, const int lda, const T *B, const int ldb,
                  const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int singleExceed4(gemmHandle<T>* handle,
                  cublasOperation_t transa, cublasOperation_t transb,
                  const int M, const int N, const int K, const T alpha,
                  const T *A, const int lda, const T *B, const int ldb,
                  const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuFit(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_nd2_1(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_np2_1(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_nd2_2(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_np2_2(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_nd2_3(gemmHandle<T>* handle,
                cublasOperation_t transa, cublasOperation_t transb,
                const int M, const int N, const int K, const T alpha,
                const T *A, const int lda, const T *B, const int ldb,
                const T beta, T *C, const int ldc);

/*--------------------------------------------------------*/

template <typename T>
cublasStatus_t cublasGemm(cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const T *alpha,
                          const T *A, int lda,
                          const T *B, int ldb,
                          const T *beta,
                          T *C, int ldc);

/*--------------------------------------------------------*/

}

/*--------------------------------------------------------*/

#endif /* _HCL_GEMMINTERNAL_H_ */

/*--------------------------------------------------------*/
