
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cstring>
#include "gemminternal.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int multigpuExceed_nd2_2(gemmHandle<T> *handle, cublasOperation_t transa,
                         cublasOperation_t transb, const int M, const int N,
                         const int K, const T alpha, const T *A, const int lda,
                         const T *B, const int ldb, const T beta, T *C,
                         const int ldc) {
  const T *hA = A;
  const T *hB = B;

  // split matrix
  int h_, w_, *height, *width, nd = handle->num_device_use;
  int rc = matrixSplitMultigpuExceed(M, N, K, handle->core_device_mem_size, nd,
                                     &h_, &w_, &height, &width);

  // initialization
  if (handle->idx_outer == 0) {
    handle->dA = (T **)malloc(sizeof(T *) * nd);
    handle->dB = (T **)malloc(sizeof(T *) * nd);
    handle->dC = (T **)malloc(sizeof(T *) * nd);

    int i, device_id = handle->device_id;
    for (i = 0; i < nd; i++) {
      // set device
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      // alloc buffers
      CUDA_SAFE_CALL(cudaMalloc((void **)(&handle->dA[i]),
                                sizeof(T) * height[0] * K * K * 2));
      CUDA_SAFE_CALL(
          cudaMalloc((void **)(&handle->dB[i]), sizeof(T) * K * width[0] * K));
      CUDA_SAFE_CALL(cudaMalloc((void **)(&handle->dC[i]),
                                sizeof(T) * height[0] * K * width[0] * K * 2));
    }
  }

  // streams and events
  cudaStream_t stream[nd][h_ * w_ / nd];
  cudaEvent_t eventForMatrixA[nd][h_ * w_ / nd];
  cudaEvent_t eventForMatrixB[nd][w_ / nd];
  cudaEvent_t eventForMatrixC[nd][h_ * w_ / nd];
  cudaEvent_t eventSend[nd][h_ * w_ / nd];
  cudaEvent_t eventRecv[nd][h_ * w_ / nd];

  int i, j, k, device_id = handle->device_id;
  for (k = 0; k < nd; k++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // create streams and events
    for (j = 0; j < w_ / nd; j++) {
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventForMatrixB[k][j],
                                              cudaEventDisableTiming));
      for (i = 0; i < h_; i++) {
        if (handle->is_overlap == 1) {
          CUDA_SAFE_CALL(cudaStreamCreate(&stream[k][i + j * h_]));
        } else {  // without overlapping
          stream[k][i + j * h_] = 0;
        }
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventForMatrixA[k][i + j * h_],
                                                cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventForMatrixC[k][i + j * h_],
                                                cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventSend[k][i + j * h_],
                                                cudaEventDisableTiming));
        CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventRecv[k][i + j * h_],
                                                cudaEventDisableTiming));
      }
    }
  }

  int ldda = K;
  int lddb = width[0] * K;
  int lddc = width[0] * K;
  T **dA = handle->dA;
  T **dB = handle->dB;
  T **dC = handle->dC;
  cublasHandle_t *handles = handle->handles;

  for (j = 0; j <= w_ - nd; j += nd) {
    for (i = 0; i <= h_ - 1; i++) {
      int idx = i + j / nd * h_;
      int buf = idx % 2;
      int buf_ = (idx + 1) % 2;
      int i_ = (i + 1) % h_;
      int j_ = (i < h_ - 1) ? j : (j + nd);

      int id;
      // initialise
      if (idx == 0) {
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          if (id > 0) {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx],
                                               eventSend[id - 1][idx], 0));
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              dB[id], sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(hB, i, j + id, height, width, N, K, 2),
              sizeof(T) * N, sizeof(T) * width[j + id] * K, K,
              cudaMemcpyHostToDevice, stream[id][idx]));
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dA[id], buf, j, height, width, N, K, 4),
              sizeof(T) * K,
              getBaddrFromIndex<T>(hA, i, j, height, width, N, K, 1),
              sizeof(T) * K, sizeof(T) * K, height[i] * K,
              cudaMemcpyHostToDevice, stream[id][idx]));
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dC[id], buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(C, i, j + id, height, width, N, K, 3),
              sizeof(T) * N, sizeof(T) * width[j + id] * K, height[i] * K,
              cudaMemcpyHostToDevice, stream[id][idx]));
          CUDA_SAFE_CALL(cudaEventRecord(eventSend[id][idx], stream[id][idx]));
        }
      }
      // gemm, then send next block of A and C, and recv block of C
      if (idx < h_ * w_ / nd - 1) {
        // gemm
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          cublasSetStream(handles[id], stream[id][idx]);
          CUBLAS_SAFE_CALL(cublasGemm<T>(
              handles[id], transb, transa, width[j + id] * K, height[i] * K, K,
              &alpha, dB[id], lddb,
              getBaddrFromIndex<T>(dA[id], buf, j, height, width, N, K, 4),
              ldda, &beta,
              getBaddrFromIndex<T>(dC[id], buf, j, height, width, N, K, 5),
              lddc));
          CUDA_SAFE_CALL(
              cudaEventRecord(eventForMatrixA[id][idx], stream[id][idx]));
          if (i == h_ - 1) {
            CUDA_SAFE_CALL(
                cudaEventRecord(eventForMatrixB[id][j / nd], stream[id][idx]));
          }
        }
        // send next block of B (possibly), A and C
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          // wait until blocks B, A, C for the previous device have been sent
          if (id == 0) {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx + 1],
                                               eventSend[nd - 1][idx], 0));
          } else {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx + 1],
                                               eventSend[id - 1][idx + 1], 0));
          }
          // if possible, send next block 0f B
          if (i == h_ - 1) {
            if (j < w_ - nd) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(
                  stream[id][idx + 1], eventForMatrixB[id][j / nd], 0));
            }
            CUDA_SAFE_CALL(cudaMemcpy2DAsync(
                dB[id], sizeof(T) * width[0] * K,
                getBaddrFromIndex<T>(hB, i, j_ + id, height, width, N, K, 2),
                sizeof(T) * N, sizeof(T) * width[j_ + id] * K, K,
                cudaMemcpyHostToDevice, stream[id][idx + 1]));
          }
          // send next block of A
          if (idx > 0) {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(
                stream[id][idx + 1], eventForMatrixA[id][idx - 1], 0));
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dA[id], buf_, j, height, width, N, K, 4),
              sizeof(T) * K,
              getBaddrFromIndex<T>(hA, i_, j, height, width, N, K, 1),
              sizeof(T) * K, sizeof(T) * K, height[i_] * K,
              cudaMemcpyHostToDevice, stream[id][idx + 1]));
          // send next block of C
          if (idx > 0) {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(
                stream[id][idx + 1], eventForMatrixB[id][idx - 1], 0));
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dC[id], buf_, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(C, i_, j_ + id, height, width, N, K, 3),
              sizeof(T) * N, sizeof(T) * width[j_ + id] * K, height[i_] * K,
              cudaMemcpyHostToDevice, stream[id][idx + 1]));
          // record the sending of blocks of this device
          CUDA_SAFE_CALL(
              cudaEventRecord(eventSend[id][idx + 1], stream[id][idx + 1]));
        }
        // recv block of C
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          if (id == 0) {
            if (idx > 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(
                  stream[id][idx], eventRecv[nd - 1][idx - 1], 0));
            }
          } else {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx],
                                               eventRecv[id - 1][idx], 0));
          }

          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(C, i, j + id, height, width, N, K, 3),
              sizeof(T) * N,
              getBaddrFromIndex<T>(dC[id], buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K, sizeof(T) * width[j + id] * K,
              height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
          CUDA_SAFE_CALL(
              cudaEventRecord(eventForMatrixB[id][idx], stream[id][idx]));
          CUDA_SAFE_CALL(cudaEventRecord(eventRecv[id][idx], stream[id][idx]));
        }
      } else {  // tail work
        // gemm
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          cublasSetStream(handles[id], stream[id][idx]);
          CUBLAS_SAFE_CALL(cublasGemm<T>(
              handles[id], transb, transa, width[j + id] * K, height[i] * K, K,
              &alpha, dB[id], lddb,
              getBaddrFromIndex<T>(dA[id], buf, j, height, width, N, K, 4),
              ldda, &beta,
              getBaddrFromIndex<T>(dC[id], buf, j, height, width, N, K, 5),
              lddc));
        }
        // recv C
        device_id = handle->device_id;
        for (id = 0; id < nd; id++) {
          CUDA_SAFE_CALL(cudaSetDevice(device_id++));
          if (id == 0) {
            if (idx > 0) {
              CUDA_SAFE_CALL(cudaStreamWaitEvent(
                  stream[id][idx], eventRecv[nd - 1][idx - 1], 0));
            }
          } else {
            CUDA_SAFE_CALL(cudaStreamWaitEvent(stream[id][idx],
                                               eventRecv[id - 1][idx], 0));
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(C, i, j + id, height, width, N, K, 3),
              sizeof(T) * N,
              getBaddrFromIndex<T>(dC[id], buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K, sizeof(T) * width[j + id] * K,
              height[i] * K, cudaMemcpyDeviceToHost, stream[id][idx]));
          CUDA_SAFE_CALL(cudaEventRecord(eventRecv[id][idx], stream[id][idx]));
        }
      }
    }
  }

  CUDA_SAFE_CALL(cudaDeviceSynchronize());

  handle->idx_outer++;

  // destroy events and streams
  device_id = handle->device_id;
  for (k = 0; k < nd; k++) {
    // set device
    CUDA_SAFE_CALL(cudaSetDevice(device_id++));
    // create streams and events
    for (j = 0; j < w_ / nd; j++) {
      CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixB[k][j]));
      for (i = 0; i < h_; i++) {
        if (handle->is_overlap == 1) {
          CUDA_SAFE_CALL(cudaStreamDestroy(stream[k][i + j * h_]));
        } else {
          stream[k][i + j * h_] = 0;
        }
        CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixA[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixC[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(eventSend[k][i + j * h_]));
        CUDA_SAFE_CALL(cudaEventDestroy(eventRecv[k][i + j * h_]));
      }
    }
  }

  // free memory
  free(height);
  height = NULL;
  free(width);
  width = NULL;
  // at the last iteration, free device memory and destroy handles
  if (handle->idx_outer == handle->num_iteration) {
    device_id = handle->device_id;
    for (j = 0; j < nd; j++) {
      CUDA_SAFE_CALL(cudaSetDevice(device_id++));
      CUDA_SAFE_CALL(cudaFree(handle->dA[j]));
      handle->dA[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(handle->dB[j]));
      handle->dB[j] = NULL;
      CUDA_SAFE_CALL(cudaFree(handle->dC[j]));
      handle->dC[j] = NULL;
    }
    free(handle->dA);
    handle->dA = NULL;
    free(handle->dB);
    handle->dB = NULL;
    free(handle->dC);
    handle->dC = NULL;
  }

  return 0;
}

/*--------------------------------------------------------*/

template int multigpuExceed_nd2_2<double>(
    gemmHandle<double> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const double alpha, const double *A, const int lda, const double *B,
    const int ldb, const double beta, double *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuExceed_nd2_2<float>(
    gemmHandle<float> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const float alpha, const float *A, const int lda, const float *B,
    const int ldb, const float beta, float *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuExceed_nd2_2<cuComplex>(
    gemmHandle<cuComplex> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const cuComplex alpha, const cuComplex *A, const int lda,
    const cuComplex *B, const int ldb, const cuComplex beta, cuComplex *C,
    const int ldc);

/*--------------------------------------------------------*/

template int multigpuExceed_nd2_2<cuDoubleComplex>(
    gemmHandle<cuDoubleComplex> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const cuDoubleComplex alpha, const cuDoubleComplex *A, const int lda,
    const cuDoubleComplex *B, const int ldb, const cuDoubleComplex beta,
    cuDoubleComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int multigpuExceed_nd2_2<__half>(
    gemmHandle<__half> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const __half alpha, const __half *A, const int lda, const __half *B,
    const int ldb, const __half beta, __half *C, const int ldc);

/*--------------------------------------------------------*/
}

/*--------------------------------------------------------*/
