
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cstring>
#include "gemminternal.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int singleExceed1(gemmHandle<T> *handle, cublasOperation_t transa,
                  cublasOperation_t transb, const int M, const int N,
                  const int K, const T alpha, const T *A, const int lda,
                  const T *B, const int ldb, const T beta, T *C,
                  const int ldc) {
  const T *hA = A;
  const T *hB = B;

  int h_, w_, *height, *width;
  int rc = matrixSplitSingleExceed(M, N, K, handle->core_device_mem_size, &h_,
                                   &w_, &height, &width);

  // initialization
  if (handle->idx_outer == 0) {
    // only one device to be used in this case, so the size of dA, dB, dC is 1
    handle->dA = (T **)malloc(sizeof(T *));
    handle->dB = (T **)malloc(sizeof(T *));
    handle->dC = (T **)malloc(sizeof(T *));

    CUDA_SAFE_CALL(cudaMalloc((void **)(&handle->dA[0]),
                              sizeof(T) * height[0] * K * K * 2));
    CUDA_SAFE_CALL(
        cudaMalloc((void **)(&handle->dB[0]), sizeof(T) * K * width[0] * K));
    CUDA_SAFE_CALL(cudaMalloc((void **)(&handle->dC[0]),
                              sizeof(T) * height[0] * K * width[0] * K * 2));
  }

  // streams and events
  cudaStream_t stream[h_ * w_];
  cudaEvent_t eventForMatrixA[h_ * w_];
  cudaEvent_t eventForMatrixB[w_];
  cudaEvent_t eventForMatrixC[h_ * w_];
  int i, j;
  for (j = 0; j < w_; j++) {
    CUDA_SAFE_CALL(
        cudaEventCreateWithFlags(&eventForMatrixB[j], cudaEventDisableTiming));
    for (i = 0; i < h_; i++) {
      if (handle->is_overlap == 1) {
        CUDA_SAFE_CALL(cudaStreamCreate(&stream[i + j * h_]));
      } else {  // without overlapping
        stream[i + j * h_] = 0;
      }
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventForMatrixA[i + j * h_],
                                              cudaEventDisableTiming));
      CUDA_SAFE_CALL(cudaEventCreateWithFlags(&eventForMatrixC[i + j * h_],
                                              cudaEventDisableTiming));
    }
  }

  int ldda = K;
  int lddb = width[0] * K;
  int lddc = width[0] * K;
  T *dA = handle->dA[0];
  T *dB = handle->dB[0];
  T *dC = handle->dC[0];

  // update in an normal order, column by column
  if (handle->idx_outer % 2 == 0) {
    // loop over send
    for (j = 0; j < w_; j++) {
      for (i = 0; i < h_; i++) {
        int idx = i + j * h_;
        int buf = idx % 2;
        // send B
        if (i == 0) {
          if (j > 0) {
            cudaStreamWaitEvent(stream[idx], eventForMatrixB[j - 1], 0);
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              dB, sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(hB, i, j, height, width, N, K, 2),
              sizeof(T) * N, sizeof(T) * width[j] * K, K,
              cudaMemcpyHostToDevice, stream[idx]));
        }
        // send A
        if (idx > 1) {
          cudaStreamWaitEvent(stream[idx], eventForMatrixA[idx - 2], 0);
        }
        CUDA_SAFE_CALL(cudaMemcpy2DAsync(
            getBaddrFromIndex<T>(dA, buf, j, height, width, N, K, 4),
            sizeof(T) * K,
            getBaddrFromIndex<T>(hA, i, j, height, width, N, K, 1),
            sizeof(T) * K, sizeof(T) * K, height[i] * K, cudaMemcpyHostToDevice,
            stream[idx]));
        // send C
        if (handle->idx_outer == 0) {
          if (idx > 1) {
            cudaStreamWaitEvent(stream[idx], eventForMatrixC[idx - 2], 0);
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
              sizeof(T) * N, sizeof(T) * width[j] * K, height[i] * K,
              cudaMemcpyHostToDevice, stream[idx]));
        } else {
          if (idx > 1) {
            cudaStreamWaitEvent(stream[idx], eventForMatrixC[idx - 2], 0);
            CUDA_SAFE_CALL(cudaMemcpy2DAsync(
                getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
                sizeof(T) * width[0] * K,
                getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
                sizeof(T) * N, sizeof(T) * width[j] * K, height[i] * K,
                cudaMemcpyHostToDevice, stream[idx]));
          }
        }

        // gemm
        cublasSetStream(handle->handles[0], stream[idx]);
        CUBLAS_SAFE_CALL(cublasGemm<T>(
            handle->handles[0], transb,
            transa, width[j] * K, height[i] * K, K, &alpha,
            dB, lddb, getBaddrFromIndex<T>(dA, buf, j, height, width, N, K, 4),
            ldda, &beta,
            getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5), lddc));
        CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixA[idx], stream[idx]));
        if (i == (h_ - 1)) {
          CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixB[j], stream[idx]));
        }

        // recv block of C
        if (handle->idx_outer == (handle->num_iteration - 1)) {
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
              sizeof(T) * N,
              getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K, sizeof(T) * width[j] * K, height[i] * K,
              cudaMemcpyDeviceToHost, stream[idx]));
          CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixC[idx], stream[idx]));
        } else {
          if (idx < (h_ * w_ - 2)) {
            CUDA_SAFE_CALL(cudaMemcpy2DAsync(
                getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
                sizeof(T) * N,
                getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
                sizeof(T) * width[0] * K, sizeof(T) * width[j] * K,
                height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
            CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixC[idx], stream[idx]));
          }
        }
      }
    }
  } else {
    // loop over send
    for (j = (w_ - 1); j >= 0; j--) {
      for (i = (h_ - 1); i >= 0; i--) {
        int idx = i + j * h_;
        int buf = idx % 2;
        // send B to device
        if (i == (h_ - 1)) {
          if (j < (w_ - 1)) {
            cudaStreamWaitEvent(stream[idx], eventForMatrixB[j], 0);
          }
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              dB, sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(hB, i, j, height, width, N, K, 2),
              sizeof(T) * N, sizeof(T) * width[j] * K, K,
              cudaMemcpyHostToDevice, stream[idx]));
        }
        // send A
        if (idx < (h_ * w_ - 2)) {
          cudaStreamWaitEvent(stream[idx], eventForMatrixA[idx + 2], 0);
        }
        CUDA_SAFE_CALL(cudaMemcpy2DAsync(
            getBaddrFromIndex<T>(dA, buf, j, height, width, N, K, 4),
            sizeof(T) * K,
            getBaddrFromIndex<T>(hA, i, j, height, width, N, K, 1),
            sizeof(T) * K, sizeof(T) * K, height[i] * K, cudaMemcpyHostToDevice,
            stream[idx]));
        // send C
        if (idx < (h_ * w_ - 2)) {
          cudaStreamWaitEvent(stream[idx], eventForMatrixC[idx + 2], 0);
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K,
              getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
              sizeof(T) * N, sizeof(T) * width[j] * K, height[i] * K,
              cudaMemcpyHostToDevice, stream[idx]));
        }

        // loop over gemm
        cublasSetStream(handle->handles[0], stream[idx]);
        CUBLAS_SAFE_CALL(cublasGemm<T>(
            handle->handles[0], transb,
            transa, width[j] * K, height[i] * K, K, &alpha,
            dB, lddb, getBaddrFromIndex<T>(dA, buf, j, height, width, N, K, 4),
            ldda, &beta,
            getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5), lddc));
        CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixA[idx], stream[idx]));
        if (i == 0) {
          CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixB[j], stream[idx]));
        }

        // recv
        if (handle->idx_outer == (handle->num_iteration - 1)) {
          CUDA_SAFE_CALL(cudaMemcpy2DAsync(
              getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
              sizeof(T) * N,
              getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
              sizeof(T) * width[0] * K, sizeof(T) * width[j] * K, height[i] * K,
              cudaMemcpyDeviceToHost, stream[idx]));
          CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixC[idx], stream[idx]));
        } else {
          if (idx > 1) {
            CUDA_SAFE_CALL(cudaMemcpy2DAsync(
                getBaddrFromIndex<T>(C, i, j, height, width, N, K, 3),
                sizeof(T) * N,
                getBaddrFromIndex<T>(dC, buf, j, height, width, N, K, 5),
                sizeof(T) * width[0] * K, sizeof(T) * width[j] * K,
                height[i] * K, cudaMemcpyDeviceToHost, stream[idx]));
            CUDA_SAFE_CALL(cudaEventRecord(eventForMatrixC[idx], stream[idx]));
          }
        }
      }
    }
  }
  CUDA_SAFE_CALL(cudaDeviceSynchronize());

  handle->idx_outer++;

  for (j = 0; j < w_; j++) {
    CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixB[j]));
    for (i = 0; i < h_; i++) {
      if (handle->is_overlap == 1) {
        CUDA_SAFE_CALL(cudaStreamDestroy(stream[i + j * h_]));
      }
      CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixA[i + j * h_]));
      CUDA_SAFE_CALL(cudaEventDestroy(eventForMatrixC[i + j * h_]));
    }
  }

  free(height);
  free(width);

  if (handle->idx_outer == handle->num_iteration) {
    CUDA_SAFE_CALL(cudaFree(dA));
    CUDA_SAFE_CALL(cudaFree(dB));
    CUDA_SAFE_CALL(cudaFree(dC));
  }

  return 0;
}

/*--------------------------------------------------------*/

template int singleExceed1<double>(gemmHandle<double> *handle,
                                   cublasOperation_t transa,
                                   cublasOperation_t transb, const int M,
                                   const int N, const int K, const double alpha,
                                   const double *A, const int lda,
                                   const double *B, const int ldb,
                                   const double beta, double *C, const int ldc);

/*--------------------------------------------------------*/

template int singleExceed1<float>(gemmHandle<float> *handle,
                                  cublasOperation_t transa,
                                  cublasOperation_t transb, const int M,
                                  const int N, const int K, const float alpha,
                                  const float *A, const int lda, const float *B,
                                  const int ldb, const float beta, float *C,
                                  const int ldc);

/*--------------------------------------------------------*/

template int singleExceed1<cuComplex>(gemmHandle<cuComplex> *handle,
                                      cublasOperation_t transa,
                                      cublasOperation_t transb, const int M,
                                      const int N, const int K,
                                      const cuComplex alpha, const cuComplex *A,
                                      const int lda, const cuComplex *B,
                                      const int ldb, const cuComplex beta,
                                      cuComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int singleExceed1<cuDoubleComplex>(
    gemmHandle<cuDoubleComplex> *handle, cublasOperation_t transa,
    cublasOperation_t transb, const int M, const int N, const int K,
    const cuDoubleComplex alpha, const cuDoubleComplex *A, const int lda,
    const cuDoubleComplex *B, const int ldb, const cuDoubleComplex beta,
    cuDoubleComplex *C, const int ldc);

/*--------------------------------------------------------*/

template int singleExceed1<__half>(gemmHandle<__half> *handle,
                                   cublasOperation_t transa,
                                   cublasOperation_t transb, const int M,
                                   const int N, const int K, const __half alpha,
                                   const __half *A, const int lda,
                                   const __half *B, const int ldb,
                                   const __half beta, __half *C, const int ldc);

/*--------------------------------------------------------*/
}

/*--------------------------------------------------------*/
