
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cuda_runtime.h>
#include <cublas_v2.h>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include "gemm.hpp"
#include "debug.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

size_t getDeviceMemory(int device) {
  if (cudaGetDevice(&device) != cudaSuccess) return 0;
  struct cudaDeviceProp properties;
  if (cudaGetDeviceProperties(&properties, device) != cudaSuccess) {
    return 0;
  }
  return properties.totalGlobalMem;
}

/*--------------------------------------------------------*/

int summer(const int* array, int size) {
  if (size == 0) {
    return 0;
  } else {
    int i = 0;
    int sum = 0;
    for (i = 0; i < size; i++) {
      sum += *(array + i);
    }
    return sum;
  }
}

/*--------------------------------------------------------*/

int matrixSplitSingleExceed(const int M, const int N, const int K,
                            const size_t coreMemSize, int* h_, int* w_,
                            int** height, int** width) {
  size_t unit = K * K;
  size_t g = (coreMemSize / unit);
  size_t h = M / K;
  size_t w = N / K;
  int _n1, _m1, n1;

  Debug::Info("coreMemSize %zu, g %zu, h %zu, w %zu.\n", coreMemSize, g, h, w);

  if ((2 + 3 * w) > g) {
    _n1 = (g - 2) / 3;
    *w_ = ceil((double)w / _n1);
  } else {
    *w_ = 1;
  }
  n1 = (w % *w_ == 0) ? (w / *w_) : (w / *w_ + 1);
  _m1 = 0.5 * ((g - n1) / (n1 + 1));
  *h_ = ceil((double)h / _m1);

  Debug::Info("h_ %d, w_ %d.\n", *h_, *w_);

  *height = (int*)malloc(sizeof(int) * (*h_));
  *width = (int*)malloc(sizeof(int) * (*w_));

  int i;
  for (i = 0; i < *h_; i++) {
    if (h % *h_ == 0) {
      (*height)[i] = h / *h_;
    } else {
      if (i < h % *h_) {
        (*height)[i] = h / *h_ + 1;
      } else {
        (*height)[i] = h / *h_;
      }
    }
  }

  for (i = 0; i < *w_; i++) {
    if (w % *w_ == 0) {
      (*width)[i] = w / *w_;
    } else {
      if (i < w % *w_) {
        (*width)[i] = w / *w_ + 1;
      } else {
        (*width)[i] = w / *w_;
      }
    }
  }

  return 0;
}

/*--------------------------------------------------------*/

int matrixSplitMultigpuExceed(const int M, const int N, const int K,
                              const size_t coreMemSize,
                              const int num_device_use, int* h_, int* w_,
                              int** height, int** width) {
  size_t unit = K * K;
  size_t g = (coreMemSize / unit);
  size_t h = M / K;
  size_t w = N / K;
  int nd = num_device_use;
  int _n1, _m1, n1;

  if ((nd * 2 + 3 * w) > nd * g) {
    _n1 = (g - 2) / 3;
    *w_ = ceil((double)w / _n1);
    if (*w_ % nd != 0) {
      *w_ = (*w_ / nd + 1) * nd;
    }
  } else {
    *w_ = nd;
  }
  n1 = (w % *w_ == 0) ? (w / *w_) : (w / *w_ + 1);
  _m1 = 0.5 * (g - n1) / (n1 + 1);
  *h_ = ceil((double)h / _m1);  // height

  *height = (int*)malloc(sizeof(int) * (*h_));
  *width = (int*)malloc(sizeof(int) * (*w_));

  int i;
  for (i = 0; i < *h_; i++) {
    if (h % *h_ == 0) {
      (*height)[i] = h / *h_;
    } else {
      if (i < h % *h_) {
        (*height)[i] = h / *h_ + 1;
      } else {
        (*height)[i] = h / *h_;
      }
    }
  }
  for (i = 0; i < *w_; i++) {
    if (w % *w_ == 0) {
      (*width)[i] = w / *w_;
    } else {
      if (i < w % *w_) {
        (*width)[i] = w / *w_ + 1;
      } else {
        (*width)[i] = w / *w_;
      }
    }
  }

  return 0;
}

/*--------------------------------------------------------*/

/*
 * Get the address of a block from index.
 */
template <typename T>
T* getBaddrFromIndex(const T* ptr, const int i, const int j, const int* height,
                     const int* width, const int N, const int K,
                     const int flag) {
  T* address = (T*)ptr;
  if (flag == 1)  // A(i)
    address += summer(height, i) * K * K;
  else if (flag == 2)  // B(i)
    address += summer(width, j) * K;
  else if (flag == 3)  // C(i,j)
    address += summer(height, i) * K * N + summer(width, j) * K;
  else if (flag == 4)  // A_(i)
    address += i * height[0] * K * K;
  else  // C_(i)
    address += i * height[0] * width[0] * K * K;

  return address;
}

/*--------------------------------------------------------*/

template <typename T>
T* getBaddr1DFromIndex(const T* ptr, const int i, const int j,
                       const int* height, const int* width, const int h_,
                       const int w_, const int K, const int flag) {
  T* address = (T*)ptr;
  if (flag == 1)  // A(i)
    address += summer(height, i) * K * K;
  else if (flag == 2)  // B(i)
    address += summer(width, j) * K * K;
  else if (flag == 3) {  // C(i,j)
    int l;
    for (l = 0; l < j; l++) {
      address += summer(height, h_) * K * width[j] * K;
    }
    address += summer(height, i) * K * width[j] * K;
  } else if (flag == 4)  // A_(i)
    address += i * height[0] * K * K;
  else  // C_(i)
    address += i * height[0] * width[0] * K * K;

  return address;
}

/*--------------------------------------------------------*/

template double* getBaddrFromIndex<double>(const double* ptr, const int i,
                                           const int j, const int* height,
                                           const int* width, const int N,
                                           const int K, const int flag);

template float* getBaddrFromIndex<float>(const float* ptr, const int i,
                                         const int j, const int* height,
                                         const int* width, const int N,
                                         const int K, const int flag);

template cuComplex* getBaddrFromIndex<cuComplex>(const cuComplex* ptr,
                                                 const int i, const int j,
                                                 const int* height,
                                                 const int* width, const int N,
                                                 const int K, const int flag);

template cuDoubleComplex* getBaddrFromIndex<cuDoubleComplex>(
    const cuDoubleComplex* ptr, const int i, const int j, const int* height,
    const int* width, const int N, const int K, const int flag);

template __half* getBaddrFromIndex<__half>(const __half* ptr, const int i,
                                           const int j, const int* height,
                                           const int* width, const int N,
                                           const int K, const int flag);

/*--------------------------------------------------------*/

template double* getBaddr1DFromIndex<double>(const double* ptr, const int i,
                                             const int j, const int* height,
                                             const int* width, const int h_,
                                             const int w_, const int K,
                                             const int flag);

template float* getBaddr1DFromIndex<float>(const float* ptr, const int i,
                                           const int j, const int* height,
                                           const int* width, const int h_,
                                           const int w_, const int K,
                                           const int flag);

template cuComplex* getBaddr1DFromIndex<cuComplex>(
    const cuComplex* ptr, const int i, const int j, const int* height,
    const int* width, const int h_, const int w_, const int K, const int flag);

template cuDoubleComplex* getBaddr1DFromIndex<cuDoubleComplex>(
    const cuDoubleComplex* ptr, const int i, const int j, const int* height,
    const int* width, const int h_, const int w_, const int K, const int flag);

template __half* getBaddr1DFromIndex<__half>(const __half* ptr, const int i,
                                             const int j, const int* height,
                                             const int* width, const int h_,
                                             const int w_, const int K,
                                             const int flag);

/*--------------------------------------------------------*/
}

/*--------------------------------------------------------*/
