
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <sys/time.h>

#include <cuda_runtime_api.h>
#include <cuda.h>
#include <cublas_v2.h>

/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

/*-----------------------------------------------------------*/

cublasHandle_t cublashandle;
double *gpuA, *gpuB, *gpuC;

/*-----------------------------------------------------------*/

int gpuinit(
    const int id,
    const int m,
    const int n,
    const int k
)
{
    cublasStatus_t status = cublasCreate(&cublashandle);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! CUBLAS initialization error\n");
        return -1;
    }

    if (cudaMalloc((void **)&gpuA, m * k * sizeof(gpuA[0])) != cudaSuccess)
    {
        fprintf(stderr, "!!!! device memory allocation error (allocate A)\n");
        return -1;
    }

    if (cudaMalloc((void **)&gpuB, k * n * sizeof(gpuB[0])) != cudaSuccess)
    {
        fprintf(stderr, "!!!! device memory allocation error (allocate B)\n");
        return -1;
    }

    if (cudaMalloc((void **)&gpuC, m * n * sizeof(gpuC[0])) != cudaSuccess)
    {
        fprintf(stderr, "!!!! device memory allocation error (allocate C)\n");
        return -1;
    }

    return 0;
}

/*-----------------------------------------------------------*/

int gpudestroy(
    const int id
)
{
    if (cudaFree(gpuA) != cudaSuccess)
    {
        fprintf(stderr, "!!!! memory free error (A)\n");
        return -1;
    }

    if (cudaFree(gpuB) != cudaSuccess)
    {
        fprintf(stderr, "!!!! memory free error (B)\n");
        return -1;
    }

    if (cudaFree(gpuC) != cudaSuccess)
    {
        fprintf(stderr, "!!!! memory free error (C)\n");
        return -1;
    }

    cublasStatus_t status = cublasDestroy(cublashandle);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! CUBLAS initialization error\n");
        return -1;
    }

    cudaError_t err = cudaDeviceReset();

    if (err != cudaSuccess)
    {
        fprintf(
           stderr,
           "Failed to deinitialize the device! error=%s\n",
           cudaGetErrorString(err));
    }

    return 0;
}

/*-----------------------------------------------------------*/

int gpudgemm(
    const int id,
    const int *myh,
    const int *myw,
    const int *kb,
    double* A,
    double* B,
    double* C,
    double* etime
)
{
    if ((*myw == 0)
        || (*myh == 0
       )
    )
    {
       *etime = 0;
       return 0;
    }

    struct timeval start, end;

    gettimeofday(&start, NULL);

    const double alpha = 1.;
    const double beta  = 1.;

    const int m = *myh;
    const int n = *myw;
    const int k = *kb;

    cublasStatus_t status;

    status = cublasSetVector(m * k, sizeof(A[0]), A, 1, gpuA, 1);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! device access error (write A)\n");
        return -1;
    }

    status = cublasSetVector(k * n, sizeof(B[0]), B, 1, gpuB, 1);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! device access error (write B)\n");
        return -1;
    }

    status = cublasSetVector(m * n, sizeof(C[0]), C, 1, gpuC, 1);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! device access error (write C)\n");
        return -1;
    }

    status = cublasDgemm(
                cublashandle,
                CUBLAS_OP_N, CUBLAS_OP_N,
                m, n, k, &alpha,
                gpuA, m, gpuB, k, &beta,
                gpuC, m
    );

    if (status != CUBLAS_STATUS_SUCCESS)
    {
       fprintf(stderr, "!!!! kernel execution error.\n");
       return -1;
    }

    status = cublasGetVector(m * n, sizeof(C[0]), gpuC, 1, C, 1);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! device access error (read C)\n");
        return -1;
    }

    gettimeofday(&end, NULL);

    double tstart = start.tv_sec + start.tv_usec/1000000.;
    double tend = end.tv_sec + end.tv_usec/1000000.;
    *etime = (tend - tstart);

    return 0;
}

/*-----------------------------------------------------------*/

unsigned int gpuNumDevices()
{
    int numDevices;
    cudaError_t result = cudaGetDeviceCount(&numDevices);
    if (result != cudaSuccess)
    {
       fprintf(stderr, "cudaGetDeviceCount error.\n");
       return -1;
    }
    return numDevices;
}


/*-----------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

/*-----------------------------------------------------------*/

int main(int argc, char** argv)
{
    if (argc != 5)
    {
       fprintf(
          stderr,
          "Usage: %s m n k numInvocations.\n",
          argv[0]);
       exit(EXIT_FAILURE);
    }

    int m = atoi(argv[1]);
    int n = atoi(argv[2]);
    int k = atoi(argv[3]);
    unsigned int numInvocations = atoi(argv[4]);

    struct timeval start, end;

    gettimeofday(&start, NULL);

    double* A = (double*)malloc(sizeof(double)*m*k);
    double* B = (double*)malloc(sizeof(double)*k*n);
    double* C = (double*)malloc(sizeof(double)*m*n);

    if ((A == NULL)
        || (B == NULL)
        || (C == NULL
       )
    )
    {
       fprintf(
         stderr,
         "Allocations of A | B | C failed...\n");
       exit(EXIT_FAILURE); 
    }

    time_t t;
    srand((unsigned)time(&t));

    int i;
    for (i = 0; i < (m*k); i++)
    {
        A[i] = rand() % (m*k);
    }

    for (i = 0; i < (k*n); i++)
    {
        B[i] = rand() % (k*n);
    }

    for (i = 0; i < (m*n); i++)
    {
        C[i] = rand() % (m*n);
    }

    gpuinit(0, m, n, k);
    gpudestroy(0);

    gettimeofday(&start, NULL);

    unsigned int invocation = 0;
    for (invocation = 0; invocation < numInvocations; invocation++)
    {
        gpuinit(0, m, n, k);

        double etime;
        gpudgemm(
            0, &m, &n, &k,
            A, B, C, &etime);

        gpudestroy(0);
    }

    gettimeofday(&end, NULL);

    double tstart = start.tv_sec + start.tv_usec/1000000.;
    double tend = end.tv_sec + end.tv_usec/1000000.;
    double t_sec = (tend - tstart) / (double)numInvocations;

    double N1 = m;
    double N2 = N1 * n * 0.001;
    double N3 = N2 * k;
    double N4 = 2.0 * N3 * 0.001;
    double speed = N4 / t_sec;

    printf(
       "m=%d, n=%d, k=%d, time(sec)=%0.6f, speed(mflops)=%0.6f\n",
       m, n, k, t_sec, speed
    );

    free(A);
    free(B);
    free(C);

    exit(EXIT_SUCCESS);
}

/*----------------------------------------------------------------------------*/

