
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <sys/time.h>

#include <cuda_runtime_api.h>
#include <cuda.h>
#include <cublasXt.h>

/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif

/*-----------------------------------------------------------*/

cublasXtHandle_t cublashandle;

/*-----------------------------------------------------------*/

int gpuinit(
    const int id,
    const int m,
    const int n,
    const int k
)
{
    cublasStatus_t status = cublasXtCreate(&cublashandle);

    int devices[1] = { id };

    if (cublasXtDeviceSelect(cublashandle, 1, devices) != CUBLAS_STATUS_SUCCESS) 
    {
       fprintf(stderr, "cublasXtDeviceSelect Failure\n"); 
       return 1;
    }

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! CUBLAS initialization error\n");
        return -1;
    }

    return 0;
}

/*-----------------------------------------------------------*/

int gpudestroy(
    const int id
)
{
    cublasStatus_t status = cublasXtDestroy(cublashandle);

    if (status != CUBLAS_STATUS_SUCCESS)
    {
        fprintf(stderr, "!!!! CUBLAS initialization error\n");
        return -1;
    }

    return 0;
}

/*-----------------------------------------------------------*/

int gpudgemm(
    const int id,
    const int *myh,
    const int *myw,
    const int *kb,
    double* A,
    double* B,
    double* C,
    double* etime
)
{
    if ((*myw == 0)
        || (*myh == 0
       )
    )
    {
       *etime = 0;
       return 0;
    }

    struct timeval start, end;

    gettimeofday(&start, NULL);

    const double alpha = 1.;
    const double beta  = 1.;

    const int m = *myh;
    const int n = *myw;
    const int k = *kb;

    cublasStatus_t status = cublasXtDgemm(
                cublashandle,
                CUBLAS_OP_N, CUBLAS_OP_N,
                m, n, k, &alpha,
                A, m, B, k, &beta,
                C, m
    );

    if (status != CUBLAS_STATUS_SUCCESS)
    {
       fprintf(stderr, "!!!! kernel execution error.\n");
       return -1;
    }

    gettimeofday(&end, NULL);

    double tstart = start.tv_sec + start.tv_usec/1000000.;
    double tend = end.tv_sec + end.tv_usec/1000000.;
    *etime = (tend - tstart);

    return 0;
}


/*-----------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

/*-----------------------------------------------------------*/

int main(int argc, char** argv)
{
    if (argc != 5)
    {
       fprintf(
          stderr,
          "Usage: %s m n k numInvocations.\n",
          argv[0]);
       exit(EXIT_FAILURE);
    }

    int m = atoi(argv[1]);
    int n = atoi(argv[2]);
    int k = atoi(argv[3]);
    unsigned int numInvocations = atoi(argv[4]);

    struct timeval start, end;

    gettimeofday(&start, NULL);

    double* A = (double*)malloc(sizeof(double)*m*k);
    double* B = (double*)malloc(sizeof(double)*k*n);
    double* C = (double*)malloc(sizeof(double)*m*n);

    if ((A == NULL)
        || (B == NULL)
        || (C == NULL
       )
    )
    {
       fprintf(
         stderr,
         "Allocations of A | B | C failed...\n");
       exit(EXIT_FAILURE); 
    }

    time_t t;
    srand((unsigned)time(&t));

    int i;
    for (i = 0; i < (m*k); i++)
    {
        A[i] = rand() % (m*k);
    }

    for (i = 0; i < (k*n); i++)
    {
        B[i] = rand() % (k*n);
    }

    for (i = 0; i < (m*n); i++)
    {
        C[i] = rand() % (m*n);
    }

    {
        double etime;

        gpuinit(0, m, n, k);

        gpudgemm(
            0, &m, &n, &k,
            A, B, C, &etime);

        gpudestroy(0);
    }

    gettimeofday(&start, NULL);

    unsigned int invocation = 0;
    for (invocation = 0; invocation < numInvocations; invocation++)
    {
        double etime;

        gpuinit(0, m, n, k);

        gpudgemm(
            0, &m, &n, &k,
            A, B, C, &etime);

        gpudestroy(0);
    }

    gettimeofday(&end, NULL);

    printf("Printing 3 random elements of C...\n");
    printf(
       "%d %lf, %d %lf, %d %lf.\n",
       rand() % (m*n),
       C[rand() % (m*n)],
       rand() % (m*n),
       C[rand() % (m*n)],
       rand() % (m*n),
       C[rand() % (m*n)]);


    double tstart = start.tv_sec + start.tv_usec/1000000.;
    double tend = end.tv_sec + end.tv_usec/1000000.;
    double t_sec = (tend - tstart) / (double)numInvocations;

    double N1 = m;
    double N2 = N1 * n * 0.001;
    double N3 = N2 * k;
    double N4 = 2.0 * N3 * 0.001;
    double speed = N4 / t_sec;

    printf(
       "m=%d, n=%d, k=%d, time(sec)=%0.6f, speed(mflops)=%0.6f\n",
       m, n, k, t_sec, speed
    );

    free(A);
    free(B);
    free(C);

    exit(EXIT_SUCCESS);
}

/*----------------------------------------------------------------------------*/

