
/*----------------------------------------------------------------------------*/

#include <cstdio>
#include <cstdlib>
#include <sys/time.h>
#include <time.h>
#include "gemm.hpp"
#include "debug.hpp"

/*----------------------------------------------------------------------------*/

template <typename T>
int testgemm(
    const int m,
    const int n,
    const int k,
    const int algorithm,
    const T alpha,
    const T beta,
    const unsigned int numInvocations
)
{
    struct timeval start, end;

    T* A = (T*)malloc(sizeof(T)*m*k);
    T* B = (T*)malloc(sizeof(T)*k*n);
    T* C = (T*)malloc(sizeof(T)*m*n);

    if ((A == NULL)
        || (B == NULL)
        || (C == NULL
       )
    )
    {
       fprintf(
         stderr,
         "Allocations of A | B | C failed...\n");
       return -1;
    }

    time_t t;
    srand((unsigned)time(&t));

    int i;
    for (i = 0; i < (m*k); i++)
    {
        A[i] = rand() % (m*k);
    }

    for (i = 0; i < (k*n); i++)
    {
        B[i] = rand() % (k*n);
    }

    for (i = 0; i < (m*n); i++)
    {
        C[i] = rand() % (m*n);
    }

    hcl::gemmHandle<T>* handle;

    hcl::create<T>(
       m, n, k,
       0, 1, 0,
       1, 1, algorithm,
       &handle
    );

    hcl::gemm<T>(
       handle,
       CUBLAS_OP_N, CUBLAS_OP_N,
       m, n, k,
       alpha,
       A, m, 
       B, k,
       beta,
       C, m);

    hcl::destroy<T>(handle);

    printf("Now timing...\n");

    gettimeofday(&start, NULL);

    unsigned int invocation = 0;
    for (invocation = 0; invocation < numInvocations; invocation++)
    {
        hcl::gemmHandle<T>* handle;

        hcl::create<T>(
           m, n, k,
           0, 1, 0,
           1, 1, 0,
           &handle
        );

        hcl::gemm<T>(
           handle,
           CUBLAS_OP_N, CUBLAS_OP_N,
           m, n, k,
           alpha,
           A, m, 
           B, k,
           beta,
           C, m);

        hcl::destroy<T>(handle);
    }

    gettimeofday(&end, NULL);

    printf("Printing 3 random elements of C...\n");
    printf(
       "%d %lf, %d %lf, %d %lf.\n",
       rand() % (m*n),
       C[rand() % (m*n)],
       rand() % (m*n),
       C[rand() % (m*n)],
       rand() % (m*n),
       C[rand() % (m*n)]);


    double tstart = start.tv_sec + start.tv_usec/1000000.;
    double tend = end.tv_sec + end.tv_usec/1000000.;
    double t_sec = (tend - tstart) / (double)numInvocations;

    double N1 = m;
    double N2 = N1 * n * 0.001;
    double N3 = N2 * k;
    double N4 = 2.0 * N3 * 0.001;
    double speed = N4 / t_sec;

    printf(
       "m=%d, n=%d, k=%d, time(sec)=%0.6f, speed(mflops)=%0.6f\n",
       m, n, k, t_sec, speed
    );

    free(A);
    free(B);
    free(C);

    printf("Test PASSED.\n");

    return 0;
}

/*----------------------------------------------------------------------------*/

int main(int argc, char** argv)
{
    if (argc != 7)
    {
       fprintf(
          stderr,
          "Usage: %s m n k algorithm numInvocations verbosity.\n",
          argv[0]);
       exit(EXIT_FAILURE);
    }

    int m = atoi(argv[1]);
    int n = atoi(argv[2]);
    int k = atoi(argv[3]);
    int algorithm = atoi(argv[4]);
    unsigned int numInvocations = atoi(argv[5]);
    unsigned int verbosity = atoi(argv[6]);

    hcl::Debug::setVerbosity(verbosity);

    printf("-----------------------------------\n");
    printf("Testing DGEMM...\n");

    testgemm<double>(
        m, n, k, algorithm, 1.0, 1.0, numInvocations
    );

    printf("-----------------------------------\n");
    printf("Testing SGEMM...\n");

    testgemm<float>(
        m, n, k, algorithm, 1.0, 1.0, numInvocations
    );

    printf("-----------------------------------\n");

    exit(EXIT_SUCCESS);
}

/*----------------------------------------------------------------------------*/

